@@ -15,21 +15,22 @@ from mash import io_util
 from mash.filesystem.filesystem import FileSystem, cd
 from mash.io_util import log, shell_ready_signal, print_shell_ready_signal, check_output
 from mash.shell import delimiters
-from mash.shell.delimiters import DEFINE_FUNCTION, IF, LEFT_ASSIGNMENT, RETURN, RIGHT_ASSIGNMENT, THEN
-from mash.shell.env import ENV, Environment, show
+from mash.shell.delimiters import comparators, DEFINE_FUNCTION, FALSE, IF, LEFT_ASSIGNMENT, RETURN, RIGHT_ASSIGNMENT, THEN, TRUE
+from mash.filesystem.scope import Scope, show
 from mash.shell.errors import ShellError, ShellPipeError
 from mash.shell.function import InlineFunction
-from mash.shell.parsing import expand_variables, expand_variables_inline, filter_comments, infer_infix_args, parse_commands
+from mash.shell.parsing import expand_variables, expand_variables_inline, filter_comments, infer_infix_args, parse_commands, quote_items
 from mash.util import for_any, has_method, identity, is_valid_method_name, omit_prefixes, removeprefix, split_prefixes, translate_terms
 
 
 confirmation_mode = False
 default_session_filename = '.shell_session.json'
 
-FALSE = ''
-TRUE = '1'
 COMMENT = '#'
+LAST_RESULTS = '_last_results'
+LAST_RESULTS_INDEX = '_last_results_index'
 INNER_SCOPE = 'inner_scope'
+ENV = 'env'
 
 Command = Callable[[Cmd, str], str]
 Types = Union[str, bool, int, float]
@@ -76,7 +77,13 @@ class BaseShell(Cmd):
         self.locals = FileSystem(scope())
         self.init_current_scope()
 
-        self.env = Environment(self.locals)
+        self.env = Scope(self.locals, ENV)
+        self.env[LAST_RESULTS] = []
+        self.env[LAST_RESULTS_INDEX] = 0
+
+        if env:
+            for k, v in env.items():
+                self.env[k] = v
 
         self.auto_save = False
         self.auto_reload = False
@@ -84,7 +91,6 @@ class BaseShell(Cmd):
         # internals
         self._do_char_method = self.none
         self._chars_allowed_for_char_method: List[str] = []
-        self._last_results = None
 
         self.set_infix_operators()
         if self.auto_reload:
@@ -100,8 +106,17 @@ class BaseShell(Cmd):
         """
         items = delimiters.python + delimiters.bash
         items.remove('=')
+        items.remove('#')
         return items
 
+    @property
+    def _last_results(self):
+        return self.env[LAST_RESULTS]
+
+    @property
+    def _last_results_index(self):
+        return self.env[LAST_RESULTS_INDEX]
+
     def set_infix_operators(self):
         # use this for infix operators, e.g. `a = 1`
         self.infix_operators = {'=': self.handle_set_env_variable}
@@ -122,30 +137,36 @@ class BaseShell(Cmd):
         self._do_char_method = method
         self._chars_allowed_for_char_method = chars
 
-    def eval(self, args: Iterable[str], quote=True) -> Types:
+    def eval(self, args: List[str], quote=True) -> Types:
         """Evaluate / run `args` and return the result.
         """
         if quote:
-            args = (shlex.quote(arg) for arg in args)
-            # args = ' '.join(shlex.quote(arg) for arg in args)
+            args = list(quote_items(args))
+
+        args = list(filter_comments(args))
 
-        args = ' '.join(filter_comments(args))
         if not args:
             return ''
 
+        if not self.is_function(args[0]):
+            args = ['echo'] + args
+
         k = '_eval_output'
-        line = f'{args} |> export {k}'
+        line = ' '.join(args)
+        line = f'{line} |> export {k}'
 
-        self.onecmd(line)
+        with enter_new_scope(self):
 
-        # verify result
-        if k not in self.env and not self._last_results:
-            raise ShellError('eval() failed')
+            self.onecmd(line)
 
-        result = self._retrieve_eval_result(k)
+            # verify result
+            if k not in self.env and not self._last_results:
+                raise ShellError('eval() failed')
 
-        if k in self.env:
-            del self.env[k]
+            result = self._retrieve_eval_result(k)
+
+            if k in self.env:
+                del self.env[k]
 
         return result
 
@@ -158,6 +179,19 @@ class BaseShell(Cmd):
 
         raise RuntimeError('Cannot retrieve result')
 
+    def eval_compare(self, line: str) -> bool:
+        f, *_ = line.split(' ')
+        terms = line.split(' ')
+
+        if self.is_function(f):
+            result = self.eval(terms)
+        elif for_any(comparators, contains, line):
+            result = self.eval(['math'] + terms)
+        else:
+            result = line
+
+        return result != FALSE
+
     def onecmd_prehook(self, line):
         """Similar to cmd.precmd but executed before cmd.onecmd
         """
@@ -175,11 +209,15 @@ class BaseShell(Cmd):
         """
         return ''
 
-    def _save_result(self, value, overwrite=True):
-        if overwrite:
-            self._last_results = [value]
-        else:
-            self._last_results.append(value)
+    def _save_result(self, value):
+        logging.debug(f'_save_result [{self._last_results_index}]: {value}')
+
+        if len(self._last_results) < self._last_results_index:
+            raise ShellError('Invalid state')
+        if len(self._last_results) == self._last_results_index:
+            self._last_results.append(None)
+
+        self._last_results[self._last_results_index] = value
 
     def is_function(self, func_name: str) -> bool:
         return has_method(self, f'do_{func_name}') or self.is_inline_function(func_name)
@@ -214,7 +252,6 @@ class BaseShell(Cmd):
                 f'Assignments cannot be used inside other assignments: {assignee}')
 
         self.locals.set(LEFT_ASSIGNMENT, keys)
-        self.set_env_variables(keys, '')
 
         # return value must be empty to prevent side-effects in the next command
         return ''
@@ -251,6 +288,9 @@ class BaseShell(Cmd):
         logging.info(f'Cmd = !{args}')
         return check_output(args)
 
+    def do_fail(self, msg: str):
+        raise ShellError(f'Fail: {msg}')
+
     def do_reduce(self, *args: str):
         """Reduce a sequence of items to using an operator.
 
@@ -315,8 +355,10 @@ class BaseShell(Cmd):
         println a b |> map echo prefix $ suffix
         ```
         """
+        self.env[LAST_RESULTS] = []
+
         lines = args.split(delimiter)
-        msg = 'Not enough arguments. Usage: `map f [args] $ [args]`.'
+        msg = 'Not enough arguments. Usage: `map f [args..] $ [args..]`.'
         if len(lines) <= 1:
             log(msg)
             return
@@ -336,7 +378,7 @@ class BaseShell(Cmd):
 
         # collect all results
         results = []
-        for line in lines:
+        for j, line in enumerate(lines):
             local_args = args.copy()
             line = line.split(' ')
 
@@ -347,8 +389,10 @@ class BaseShell(Cmd):
 
             line = [f] + local_args
 
+            self.env[LAST_RESULTS_INDEX] = j
             results.append(self.run_single_command(line))
 
+        self.env[LAST_RESULTS_INDEX] = 0
         return delimiter.join([str(result) for result in results])
 
     def do_foreach(self, args):
@@ -369,6 +413,11 @@ class BaseShell(Cmd):
         """
         return '\n'.join(args.split(' '))
 
+    def do_strip(self, args: str) -> str:
+        """Convert a space-separated string to a newline-separates string.
+        """
+        return args.strip()
+
     def do_int(self, args: str) -> str:
         self._save_result(int(args))
         return ''
@@ -378,13 +427,13 @@ class BaseShell(Cmd):
         return ''
 
     def do_bool(self, args: str) -> str:
-        self._save_result(bool(args))
+        self._save_result(args != FALSE)
         return ''
 
     def do_not(self, args: str) -> str:
-        if args:
-            return FALSE
-        return TRUE
+        if args == FALSE:
+            return TRUE
+        return FALSE
 
     ############################################################################
     # Overrides
@@ -446,7 +495,6 @@ class BaseShell(Cmd):
                 self.reset_locals()
                 raise
 
-            self.reset_locals()
             return result
 
         if line in self._chars_allowed_for_char_method:
@@ -469,6 +517,8 @@ class BaseShell(Cmd):
             return
 
         self.locals.set(IF, [])
+        if LEFT_ASSIGNMENT in self.locals:
+            self.locals.rm(LEFT_ASSIGNMENT)
 
         for i, line in enumerate(lines):
 
@@ -501,9 +551,12 @@ class BaseShell(Cmd):
             # handle inline `<-`
             if DEFINE_FUNCTION not in self.locals \
                     and LEFT_ASSIGNMENT in self.locals \
-                    and 'assign' not in line:
+                    and 'assign' not in line \
+                    and not \
+                        (len(lines) > i + 1
+                         and lines[i+1][0] == LEFT_ASSIGNMENT):
                 self._save_assignee(result)
-                result = ''
+                result = None
 
         if DEFINE_FUNCTION in self.locals:
             if not self.locals[DEFINE_FUNCTION].multiline:
@@ -565,6 +618,9 @@ class BaseShell(Cmd):
                     return ''
                 # otherwise continue
 
+                if not self.is_function(line.split(' ')[0]):
+                    line = 'echo ' + line
+
             if prefixes[-1] in delimiters.bash:
                 return self.pipe_cmd_sh(line, prev_result, delimiter=prefixes[-1])
 
@@ -588,9 +644,10 @@ class BaseShell(Cmd):
                             f'The previous if-then statement was not closed')
 
                 if self.locals[IF] and not self.locals[IF][-1]['value']:
+                    # force value to be false when previous conditions are false
                     value = False
                 else:
-                    value = line != ''
+                    value = self.eval_compare(line)
 
                 self.locals[IF].append({'value': value, 'depth': 0})
                 return ''
@@ -609,8 +666,9 @@ class BaseShell(Cmd):
 
         if result is None:
             raise ShellError(f'Missing return value in assignment: {keys}')
-        elif result == '' and self._last_results:
-            result = self._last_results.pop()
+        elif result.strip() == '' and self._last_results:
+            result = self._last_results
+            self.env[LAST_RESULTS] = []
 
         self.set_env_variables(keys, result)
 
@@ -668,6 +726,9 @@ class BaseShell(Cmd):
         """
         assert delimiter in delimiters.bash
 
+        if delimiter == '>-':
+            delimiter = '>'
+
         # pass last result to stdin
         line = f'echo {shlex.quote(prev_result)} {delimiter} {line}'
 
@@ -728,8 +789,6 @@ class BaseShell(Cmd):
         if result is None:
             raise ShellError(f'Missing return value in assignment: {keys}')
 
-        # TODO set multiple keys based on pattern matching
-        # e.g. i j = range(2)
         if isinstance(keys, str):
             keys = keys.split(' ')
 
@@ -759,11 +818,7 @@ class BaseShell(Cmd):
                 f'Cannot assign values to all keys: {" ".join(keys)}')
 
     def handle_set_env_variable(self, lhs: Tuple[str], *rhs: str) -> str:
-        if len(lhs) != 1:
-            raise ShellError()
-
-        k = lhs[0]
-        self.set_env_variable(k, *rhs)
+        self.set_env_variables(lhs, ' '.join(rhs))
         return ''
 
     ############################################################################
@@ -778,7 +833,7 @@ class BaseShell(Cmd):
             return
 
         self.reset_locals()
-        env = self.locals[ENV]
+        env = filter_private_keys(self.locals[ENV])
 
         with open(session, 'w') as f:
             try:
@@ -818,7 +873,8 @@ class BaseShell(Cmd):
 
         self.reset_locals()
         log(f'Using session: {session}')
-        show(env)
+        print('Env')
+        show(env, when=is_public)
 
         # TODO handle key conflicts
         self.env.update(env)
@@ -836,25 +892,25 @@ class BaseShell(Cmd):
             raise ShellError(
                 f'Invalid number of arguments: {len(f.args)} arguments expected .')
 
+        # translate variables in inline functions
         for i, k in enumerate(f.args):
             # quote item to preserve `\n`
             translations[k] = shlex.quote(args[i])
 
         with enter_new_scope(self):
 
-            for line in f.inner:
-                terms = [term for term in line.split(' ') if term != '']
-                terms = list(translate_terms(terms, translations))
+            for i, k in enumerate(f.args):
+                # quote item to preserve `\n`
+                self.env[k] = shlex.quote(args[i])
 
-                self.onecmd(' '.join(terms), print_result=False)
+            for line in f.inner:
+                self.onecmd(line, print_result=False)
 
             terms = [term for term in f.command.split(' ') if term != '']
-            terms = list(translate_terms(terms, translations))
-
-            first_func = terms[0]
-            if not self.is_function(first_func):
-                terms = ['print'] + terms
+            if not f.multiline:
+                terms = list(translate_terms(terms, translations))
 
+            # don't re-quote terms to maintain newlines
             result = self.eval(terms, quote=False)
 
         return result
@@ -926,6 +982,22 @@ def is_function_definition(terms: List[str]) -> bool:
     return first.startswith('(') and last.endswith(')')
 
 
+def is_public(key: str) -> bool:
+    return not is_private(key)
+
+
+def is_private(key: str) -> bool:
+    return key.startswith('_')
+
+
+def filter_private_keys(env: dict) -> dict:
+    env = env.copy()
+    for k in list(env.keys()):
+        if is_private(k):
+            del env[k]
+    return env
+
+
 def scope() -> dict:
     return defaultdict(dict)
 
