@@ -1,23 +1,31 @@
 from enum import Enum
 
+# language constants to a new file
+# TODO mv these to a new module
+FALSE = ''
+TRUE = '1'
+
+comparators = ['==', '!=', '>', '<']
+
 
 class Python(Enum):
     RETURN = 'return'
+    NEW_COMMAND = ';'
+    LEFT_ASSIGNMENT = '<-'
     IF = 'if'
     THEN = 'then'
-    NEW_COMMAND = ';'
     DEFINE_FUNCTION = ':'
     RIGHT_ASSIGNMENT = '->'
-    LEFT_ASSIGNMENT = '<-'
     PIPE = '|>'
     MAP = '>>='
     SET_ENV_VARIABLE = '='
+    INLINE_COMMENT = '#'
     # ELSE = 'else'
     # AND = 'and'
     # OR = 'or'
 
 
-bash = ['|', '>', '>>', '1>', '1>>', '2>', '2>>']
+bash = ['|', '>-', '>>', '1>', '1>>', '2>', '2>>']
 python = [o.value for o in Python]
 all = bash + python
 
